from typing import Any, Callable, Dict, Tuple, TypeVar
from instaui.vars.ref import Ref
from instaui.vars._types import InputBindingType, OutputSetType
from instaui.vars.mixin_types.py_binding import CanInputMixin, CanOutputMixin
from instaui.vars.mixin_types.observable import ObservableMixin
from instaui.vars.mixin_types.element_binding import ElementBindingMixin
from instaui.vars.mixin_types.pathable import CanPathPropMixin
from instaui.vars.mixin_types.str_format_binding import StrFormatBindingMixin
from instaui.common.jsonable import Jsonable
from instaui.vars.ref_base import RefBase


from pydantic import BaseModel, RootModel


_T = TypeVar("_T")


_ProxyModel = RootModel


class RefProxy(
    CanInputMixin,
    ObservableMixin,
    CanOutputMixin,
    CanPathPropMixin,
    StrFormatBindingMixin,
    ElementBindingMixin,
    Jsonable,
):
    def __init__(self, data, ref_base: RefBase) -> None:
        self._ref_ = ref_base
        self._prop_names_ = set(data.keys()) if isinstance(data, dict) else set()

    @property
    def __ref_(self):
        return super().__getattribute__("_ref_")

    def __getattribute__(self, name):
        if name not in super().__getattribute__("_prop_names_"):
            return super().__getattribute__(name)

        return self.__ref_[name]

    def __getitem__(self, name):
        return self.__ref_[name]

    def inverse(self):
        return self.__ref_.inverse()

    def __add__(self, other: str):
        return self.__ref_ + other

    def __radd__(self, other: str):
        return other + self.__ref_

    def _to_element_binding_config(self) -> Dict:
        return self.__ref_._to_element_binding_config()

    def _to_input_config(self):
        return self.__ref_._to_input_config()

    def _to_observable_config(self):
        return self.__ref_._to_observable_config()

    def _to_path_prop_binding_config(self) -> Dict:
        return self.__ref_._to_path_prop_binding_config()

    def _to_output_config(self):
        return self.__ref_._to_output_config()

    def _to_str_format_binding(self, order: int) -> Tuple[str, str]:
        return self.__ref_._to_str_format_binding(order)

    def _to_json_dict(self):
        return self.__ref_._to_json_dict()

    def _to_event_output_type(self) -> OutputSetType:
        return self.__ref_._to_event_output_type()

    def _to_event_input_type(self) -> InputBindingType:
        return self.__ref_._to_event_input_type()


class StateModel(BaseModel, Jsonable):
    pass

    def _to_json_dict(self):
        return self.model_dump()


def state(
    value: _T,
    deep_compare: bool = False,
) -> _T:
    """
    Creates a reactive state object that tracks changes and notifies dependencies.

    Args:
        value (_T): The initial value to wrap in a reactive state container.
                   Can be any type (primitive, object, or complex data structure).

    # Example:
    .. code-block:: python
        from instaui import ui,html
        count = ui.state(0)

        html.number(count)
        ui.text(count)
    """

    create_proxy, is_proxy = _ready_state(value)
    if is_proxy:
        return value

    return create_proxy(lambda value: Ref(value, deep_compare=deep_compare))


def _ready_state(value) -> Tuple[Callable, bool]:
    """
    .. code-block:: python
        create_proxy, is_proxy = _ready_state(value)
        if is_proxy:
            return value

        return create_proxy(lambda value: Ref(value, deep_compare=deep_compare))
    """
    if isinstance(value, RefProxy):
        return value, True  # type: ignore

    def create_proxy(builder: Callable[[Any], RefBase]):
        new_value = _ProxyModel(value).model_dump()
        obj = RefProxy(new_value, builder(new_value))
        return obj

    return create_proxy, False



from __future__ import annotations
from typing import Optional

from .ref_base import RefBase
from .state import _ready_state


class UseDarkRef(RefBase):
    def __init__(
        self,
        deep_compare: bool = False,
        options: Optional[dict] = None,
    ) -> None:
        super().__init__(ref_type="useDark", deep_compare=deep_compare, args=options)


def use_dark(*, storage_key: Optional[str] = None, deep_compare: bool = False) -> bool:
    """
    On start up, it reads the value from localStorage/sessionStorage (the key is configurable) to see if there is a user configured color scheme, if not, it will use users' system preferences.

    Args:
        storage_key (Optional[str], optional): The local storage key to read the value from. Defaults to None(eg. "insta-color-scheme").

    Example:
    .. code-block:: python
        from instaui import ui,html

        @ui.page('/')
        def index():
            dark = ui.use_dark()
            html.checkbox(dark)
    """

    create_proxy, _ = _ready_state(None)

    return create_proxy(
        lambda _: UseDarkRef(
            deep_compare=deep_compare, options={"storageKey": storage_key}
        )
    )



from typing import Mapping
from instaui.vars.mixin_types.element_binding import ElementBindingMixin


def gen_responsive_props(props: Mapping):
    """
    def __init__(
        self,
        **kwargs: Unpack[TFlexProps],
    ):
        super().__init__("flex")

        self.props(gen_responsive_props(kwargs))
    """
    binding_props = {}
    static_props = {}

    for key, value in props.items():
        if value is None:
            continue
        if isinstance(value, ElementBindingMixin):
            binding_props[key] = value._to_element_binding_config()
        else:
            static_props[key] = value

    layout_props = {}

    if static_props:
        layout_props["props"] = static_props

    if binding_props:
        layout_props["bind"] = binding_props

    if not layout_props:
        return {}

    return {"_resp": layout_props}



from __future__ import annotations
from typing import Any, Dict, Optional
from instaui.common.jsonable import Jsonable
from instaui.runtime._app import get_current_scope
from instaui.vars._types import InputBindingType, OutputSetType
from instaui.vars.path_var import PathVar
from instaui.missing import MISSING

from .mixin_types.var_type import VarMixin
from .mixin_types.py_binding import CanInputMixin, CanOutputMixin
from .mixin_types.observable import ObservableMixin
from .mixin_types.element_binding import ElementBindingMixin
from .mixin_types.pathable import CanPathPropMixin
from .mixin_types.str_format_binding import StrFormatBindingMixin


class RefBase(
    Jsonable,
    PathVar,
    VarMixin,
    ObservableMixin,
    CanInputMixin,
    CanOutputMixin,
    CanPathPropMixin,
    StrFormatBindingMixin,
    ElementBindingMixin,
):
    def __init__(
        self,
        *,
        ref_type: Optional[str] = None,
        value: Optional[Any] = MISSING,
        deep_compare: bool = False,
        args: Optional[Dict] = None,
    ) -> None:
        scope = get_current_scope()
        self.__register_info = scope.register_ref_task(self)
        self._deep_compare = deep_compare
        self._value = value
        self._args = args
        self._ref_type = ref_type

    def __to_binding_config(self):
        return {
            "id": self.__register_info.var_id,
            "sid": self.__register_info.scope_id,
        }

    def _to_pathable_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_path_prop_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_observable_config(self):
        return self.__to_binding_config()

    def _to_input_config(self):
        return self.__to_binding_config()

    def _to_output_config(self):
        return self.__to_binding_config()

    def _to_element_binding_config(self):
        return self.__to_binding_config()

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["id"] = self.__register_info.var_id

        if self._deep_compare is True:
            data["deepCompare"] = True

        if self._args:
            data["args"] = {k: v for k, v in self._args.items() if v is not None}

        if self._ref_type:
            data["type"] = self._ref_type

        if self._value is not MISSING:
            data["value"] = self._value

        return data

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref

    def _to_event_output_type(self) -> OutputSetType:
        return OutputSetType.Ref
